#include <sourcemod>
#include <sdktools>
#include <sdkhooks>
#include <colors>

#define PLUGIN_VERSION "1.1"

#define PREFIX "{green}[NoBlock]{default} "

/* Cvars Handles*/
new Handle:g_hCVEnable;
new Handle:g_hCVAnnounce;
new Handle:g_hCVPlayer;
new Handle:g_hCVTime;
new Handle:g_hCVPlayerSpawn;
new Handle:g_hCVTimeSpawn;
new Handle:g_hCVNades;

/* Timer Handles */
new Handle:g_hSpawnNoBlockTimer = INVALID_HANDLE;
new Handle:g_hNoBlockTimer = INVALID_HANDLE;
new Handle:g_hStuckTimer[MAXPLAYERS+1];

new g_ioffsCollisionGroup;
new g_iRoundStartTime = -1;

/* Plugin Info */
public Plugin:myinfo = 
{
	name = "SpawnNoBlock",
	author = "fluxX",
	description = "Removes player collisions.",
	version = PLUGIN_VERSION,
	url = "http://www.wcfan.de"
};

public OnPluginStart()
{
	// Convars
	new Handle:hVersion = CreateConVar("sm_spawnnoblock_version", PLUGIN_VERSION, "SpawnNoblock Version", FCVAR_PLUGIN | FCVAR_SPONLY | FCVAR_REPLICATED | FCVAR_NOTIFY | FCVAR_DONTRECORD);
	if(hVersion != INVALID_HANDLE)
		SetConVarString(hVersion, PLUGIN_VERSION);
	
	g_hCVEnable = CreateConVar("sm_spawnnoblock_enabled", "1", "Plugin SpawnNoBlock (0 = Disabled, 1 = Enabled).", FCVAR_PLUGIN);
	g_hCVAnnounce = CreateConVar("sm_spawnnoblock_announce", "180", "Time between the first Annouce and the second (0 = Disabled).", FCVAR_PLUGIN);
	g_hCVPlayer = CreateConVar("sm_noblock_player", "1", "Player noblock by typing !noblock/!nb (0 = Disabled, 1 = Enabled).", FCVAR_PLUGIN);
	g_hCVTime = CreateConVar("sm_noblock_time", "10.0", "How long will be players unblockable by typing !noblock/!nb?", FCVAR_PLUGIN, true, 1.0, true, 60.0);
	g_hCVPlayerSpawn = CreateConVar("sm_spawnnoblock_player", "1", "Player noblock on spawn (0 = Disabled, 1 = Enabled).", FCVAR_PLUGIN);
	g_hCVTimeSpawn = CreateConVar("sm_spawnnoblock_time", "10.0", "How long will be players unblockable on spawn?", FCVAR_PLUGIN, true, 1.0, true, 60.0);
	g_hCVNades = CreateConVar("sm_spawnnoblock_nades", "0", "Enable noblock on nades (0 = Disabled, 1 = Enabled).", FCVAR_PLUGIN);
	
	HookConVarChange(g_hCVEnable, OnConVarChange);
	HookConVarChange(g_hCVAnnounce, OnConVarChange);
	HookConVarChange(g_hCVPlayer, OnConVarChange);
	HookConVarChange(g_hCVTime, OnConVarChange);
	HookConVarChange(g_hCVPlayerSpawn, OnConVarChange);
	HookConVarChange(g_hCVTimeSpawn, OnConVarChange);
	HookConVarChange(g_hCVNades, OnConVarChange);	
	
	// Create the autoconfig
	AutoExecConfig(true, "spawnnoblock");
	
	// Load Translation
	LoadTranslations("spawnnoblock.phrases");
		
	HookEvent("player_spawn", Event_PlayerSpawn);
	HookEvent("round_freeze_end", Event_RoundStart);
	HookEvent("round_end", Event_RoundEnd);
	
	// Commands
	RegConsoleCmd("say", Cmd_Noblock, "Enable Noblock");
	RegConsoleCmd("say_team", Cmd_Noblock, "Enable Noblock");
	
	g_ioffsCollisionGroup = FindSendPropOffs("CBaseEntity", "m_CollisionGroup");
	if (g_ioffsCollisionGroup == -1)
	{
		SetConVarInt(g_hCVEnable,0);
		PrintToServer("* FATAL ERROR: Failed to get offset for CBaseEntity::m_CollisionGroup");
	}
	else
	{
		SetConVarInt(g_hCVEnable,1);
	}
}

public OnMapStart()
{
	if(GetConVarInt(g_hCVEnable) == 1 && GetConVarInt(g_hCVAnnounce) > 0)
	{
		CreateTimer(GetConVarFloat(g_hCVAnnounce), AnnounceMsg, _, TIMER_REPEAT|TIMER_FLAG_NO_MAPCHANGE);
	}
}

public OnMapEnd()
{
	g_iRoundStartTime = -1;
	
	// Kill Timers	
	ClearHandle(g_hNoBlockTimer);
	ClearHandle(g_hSpawnNoBlockTimer);
}

public OnClientDisconnect(client)
{
	ClearHandle(g_hStuckTimer[client]);
}

public OnConVarChange(Handle:hCvar, const String:oldValue[], const String:newValue[])
{
	if(hCvar == g_hCVEnable)
		SetConVarString(g_hCVEnable, newValue);
	if(hCvar == g_hCVAnnounce)
		SetConVarString(g_hCVAnnounce, newValue);
	if(hCvar == g_hCVPlayer)
		SetConVarString(g_hCVPlayer, newValue);
	if(hCvar == g_hCVTime)
		SetConVarString(g_hCVTime, newValue);
	if(hCvar == g_hCVPlayerSpawn)
		SetConVarString(g_hCVPlayerSpawn, newValue);
	if(hCvar == g_hCVTimeSpawn)
		SetConVarString(g_hCVTimeSpawn, newValue);
	if(hCvar == g_hCVNades)
		SetConVarString(g_hCVNades, newValue);
}

public Event_RoundStart(Handle:event, const String:name[], bool:dontBroadcast)
{
	g_iRoundStartTime = -1;
	
	// Kill Timers
	ClearHandle(g_hNoBlockTimer);
	ClearHandle(g_hSpawnNoBlockTimer);
	
	for(new i=1; i <= MaxClients; i++)
		ClearHandle(g_hStuckTimer[i]);
	
	if(GetConVarInt(g_hCVEnable) == 1)
	{
		g_iRoundStartTime = GetTime();
		g_hSpawnNoBlockTimer = CreateTimer(GetConVarFloat(g_hCVTimeSpawn),SetBlockTimerSpawn, _, TIMER_FLAG_NO_MAPCHANGE);
	}
	
}

public Event_RoundEnd(Handle:event, const String:name[], bool:dontBroadcast)
{
	g_iRoundStartTime = -1;
	
	// Kill Timers	
	ClearHandle(g_hNoBlockTimer);
	ClearHandle(g_hSpawnNoBlockTimer);
}

// Enable NoBlock on players on playerspawn
public Event_PlayerSpawn(Handle:event, const String:name[], bool:dontBroadcast)
{
	new userid = GetEventInt(event, "userid");
	new client = GetClientOfUserId(userid);
	
	if(GetConVarInt(g_hCVEnable) == 1 && GetConVarInt(g_hCVPlayerSpawn) == 1 && ((GetTime() - g_iRoundStartTime) < GetConVarInt(g_hCVTimeSpawn) || g_iRoundStartTime == -1)	&& IsPlayerAlive(client) && GetClientTeam(client) > 1)
	{
		// Disable Blocking on the Client
		SetNoBlock(client);
		CPrintToChat(client,"%s%t", PREFIX, "SpawnNoblock enabled", (g_iRoundStartTime!=-1?(GetConVarInt(g_hCVTimeSpawn) - GetTime() + g_iRoundStartTime):GetConVarInt(g_hCVTimeSpawn)));
	}
}

public Action:Cmd_Noblock(client, args)
{
	new String:line[32];
	new String:name[MAX_NAME_LENGTH];
	
	if (args > 0)
	{		
		if(GetConVarInt(g_hCVEnable) == 1 && GetConVarInt(g_hCVPlayer) == 1 && IsPlayerAlive(client))
		{
			GetCmdArg(1,line,sizeof(line));	
			
			if (StrEqual(line, "!noblock", false) || StrEqual(line, "!nb", false))
			{
				//Check if the player is blockable
				if(!IsNoBlockable(client))
				{
					for(new i = 1; i <= MaxClients; i++)
					{
						if(IsClientInGame(i) && IsPlayerAlive(i))
						{
							// Disable Blocking on the Client
							SetNoBlock(i);
						}
					}
					GetClientName(client, name, sizeof(name));
					CPrintToChatAll("%s%t", PREFIX, "Noblock enabled", GetConVarInt(g_hCVTime), name);
					g_hNoBlockTimer = CreateTimer(GetConVarFloat(g_hCVTime), SetBlockTimer, TIMER_FLAG_NO_MAPCHANGE);
				}
				else
				{
					CPrintToChat(client,"%s%t", PREFIX, "Already noblockable");
				}
			}
		}
	}
	return Plugin_Continue;
}

//To make sure the player don't stuck in another player
stock bool:IsPlayerStuck(client)
{
	decl Float:vOrigin[3], Float:vMins[3], Float:vMaxs[3];
	GetClientAbsOrigin(client, vOrigin);
	GetEntPropVector(client, Prop_Send, "m_vecMins", vMins);
	GetEntPropVector(client, Prop_Send, "m_vecMaxs", vMaxs);
	
	TR_TraceHullFilter(vOrigin, vOrigin, vMins, vMaxs, MASK_ALL, FilterOnlyPlayers, client);
	return TR_DidHit();
}

public bool:FilterOnlyPlayers(entity, contentsMask, any:data)
{
	if(entity != data && entity > 0 && entity <= MaxClients) 
    	return true;
	return false;
}

// Set Noblock on nades
public OnEntityCreated(entity, const String:classname[])
{
	if(GetConVarInt(g_hCVEnable) == 1 && GetConVarInt(g_hCVNades) == 1)
	{
		if (StrEqual(classname, "hegrenade_projectile"))
		{
			SetNoBlock(entity);
		} else if (StrEqual(classname, "flashbang_projectile"))
		{
			SetNoBlock(entity);
		} else if (StrEqual(classname, "smokegrenade_projectile"))
		{
			SetNoBlock(entity);
		}
	}
}

// Player Noblocking Expires
public Action:SetBlockTimerSpawn(Handle:timer, any:userid)
{
	g_hSpawnNoBlockTimer = INVALID_HANDLE;
	
	if(GetConVarInt(g_hCVEnable) == 1 && GetConVarInt(g_hCVPlayerSpawn) == 1)
	{
		for(new i = 1; i <= MaxClients; i++)
		{
			if(IsClientInGame(i) && IsPlayerAlive(i))
			{
				// Check if the player stuck in another player
				if(!IsPlayerStuck(i))
				{
					// If no -> Enable Blocking on the Client
					SetBlock(i);
				} 
				else
				{
					// If yes -> Create a timer and check in 2 secods again
					g_hStuckTimer[i] = CreateTimer(2.0, SetStuckBlock, GetClientUserId(i), TIMER_FLAG_NO_MAPCHANGE);					
				}
			}
		}
		CPrintToChatAll("%s%t", PREFIX, "SpawnNoblock disabled");
	}
	return Plugin_Stop;
}

// Player Noblocking Expires
public Action:SetBlockTimer(Handle:timer, any:userid)
{
	g_hNoBlockTimer = INVALID_HANDLE;
	
	if(GetConVarInt(g_hCVEnable) == 1 && GetConVarInt(g_hCVPlayer) == 1)
	{
		for(new i = 1; i <=MaxClients; i++)
		{
			if(IsClientInGame(i) && IsPlayerAlive(i))
			{
				// Check if the player stuck in another player
				if(!IsPlayerStuck(i))
				{
					// If no -> Enable Blocking on the Client
					SetBlock(i);
				} 
				else
				{
					// If yes -> Create a timer and check in 2 secods again
					g_hStuckTimer[i] = CreateTimer(2.0, SetStuckBlock, GetClientUserId(i), TIMER_FLAG_NO_MAPCHANGE);					
				}
			}
		}
		CPrintToChatAll("%s%t", PREFIX, "Noblock disabled");
	}
	return Plugin_Stop;
}

public Action:SetStuckBlock(Handle:timer, any:userid)
{
	new client = GetClientOfUserId(userid);
	
	g_hStuckTimer[client] = INVALID_HANDLE;
	
	if(GetConVarInt(g_hCVEnable) == 1 && GetConVarInt(g_hCVPlayer) == 1 || GetConVarInt(g_hCVPlayerSpawn) == 1)
	{
		if(IsClientInGame(client) && IsPlayerAlive(client))
		{
			// Check again if the player stuck in another player
			if(!IsPlayerStuck(client))
				// If no -> Enable Blocking on the Client
				SetBlock(client);
			else
				// If yes -> check again with a Timer
				g_hStuckTimer[client] = CreateTimer(2.0, SetStuckBlock, userid, TIMER_FLAG_NO_MAPCHANGE);
		}
	}
	return Plugin_Stop;
}

public Action:AnnounceMsg(Handle:timer, any:userid)
{
	CPrintToChatAll("%s%t", PREFIX, "Announce");
	return Plugin_Continue;
}

// For Noblock
public SetNoBlock(entity)
{
	SetEntData(entity, g_ioffsCollisionGroup, 2, 4, true);
}

// For Block
public SetBlock(entity)
{
	SetEntData(entity, g_ioffsCollisionGroup, 5, 4, true);
}

// Check if the player is in Block mode
public bool:IsNoBlockable(client)
{
	return GetEntData(client,g_ioffsCollisionGroup,4) == 2;
}

stock ClearHandle(&Handle:hndl)
{
	if(hndl != INVALID_HANDLE)
		CloseHandle(hndl);
	hndl = INVALID_HANDLE;
}